/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.commons.text;

/**
 * A predicate for selecting code points. Implementations of this interface must
 * be thread safe.
 *
 * @since 1.0
 */
public interface CharacterPredicate {

    /**
     * Tests the code point with this predicate.
     *
     * @param codePoint
     *            the code point to test
     * @return {@code true} if the code point matches the predicate,
     *         {@code false} otherwise
     * @since 1.0
     */
    boolean test(int codePoint);
}
